2019-05-13 09:16:40

by Yuyang Du

[permalink] [raw]
Subject: [PATCH 01/17] locking/lockdep: Add lock type enum to explicitly specify read or write locks

Add an enum to formalize lock types, as those type values now matter. No
functional change.

Signed-off-by: Yuyang Du <[email protected]>
---
include/linux/lockdep.h | 28 ++++++++++++++++++++++------
kernel/locking/lockdep.c | 23 +++++++++++++----------
2 files changed, 35 insertions(+), 16 deletions(-)

diff --git a/kernel/locking/lockdep.c b/kernel/locking/lockdep.c
index 7275d6c..e9eafcf 100644
--- a/kernel/locking/lockdep.c
+++ b/kernel/locking/lockdep.c
@@ -2297,7 +2297,10 @@ static inline void inc_chains(void)
* (Note that this has to be done separately, because the graph cannot
* detect such classes of deadlocks.)
*
- * Returns: 0 on deadlock detected, 1 on OK, 2 on recursive read
+ * Returns:
+ * 0: on deadlock detected;
+ * 1: on OK;
+ * 2: LOCK_TYPE_RECURSIVE on recursive read
*/
static int
check_deadlock(struct task_struct *curr, struct held_lock *next)
@@ -2319,15 +2322,15 @@ static inline void inc_chains(void)
* Allow read-after-read recursion of the same
* lock class (i.e. read_lock(lock)+read_lock(lock)):
*/
- if ((next->read == 2) && prev->read)
- return 2;
+ if ((next->read == LOCK_TYPE_RECURSIVE) && prev->read)
+ return LOCK_TYPE_RECURSIVE;

/*
* We're holding the nest_lock, which serializes this lock's
* nesting behaviour.
*/
if (nest)
- return 2;
+ return LOCK_TYPE_RECURSIVE;

print_deadlock_bug(curr, prev, next);
return 0;
@@ -2407,7 +2410,7 @@ static inline void inc_chains(void)
* write-lock never takes any other locks, then the reads are
* equivalent to a NOP.
*/
- if (next->read == 2 || prev->read == 2)
+ if (next->read == LOCK_TYPE_RECURSIVE || prev->read == LOCK_TYPE_RECURSIVE)
return 1;
/*
* Is the <prev> -> <next> dependency already present?
@@ -2493,7 +2496,7 @@ static inline void inc_chains(void)
* Only non-recursive-read entries get new dependencies
* added:
*/
- if (hlock->read != 2 && hlock->check) {
+ if (hlock->read != LOCK_TYPE_RECURSIVE && hlock->check) {
int ret = check_prev_add(curr, hlock, next, distance,
&trace);
if (!ret)
@@ -2877,13 +2880,13 @@ static int validate_chain(struct task_struct *curr,
* building dependencies (just like we jump over
* trylock entries):
*/
- if (ret == 2)
- hlock->read = 2;
+ if (ret == LOCK_TYPE_RECURSIVE)
+ hlock->read = LOCK_TYPE_RECURSIVE;
/*
* Add dependency only if this lock is not the head
* of the chain, and if it's not a secondary read-lock:
*/
- if (!chain_head && ret != 2) {
+ if (!chain_head && ret != LOCK_TYPE_RECURSIVE) {
if (!check_prevs_add(curr, hlock))
return 0;
}
@@ -4105,7 +4108,7 @@ static int __lock_downgrade(struct lockdep_map *lock, unsigned long ip)
curr->curr_chain_key = hlock->prev_chain_key;

WARN(hlock->read, "downgrading a read lock");
- hlock->read = 1;
+ hlock->read = LOCK_TYPE_READ;
hlock->acquire_ip = ip;

if (reacquire_held_locks(curr, depth, i))
diff --git a/include/linux/lockdep.h b/include/linux/lockdep.h
index 7c2fefa..441288c 100644
--- a/include/linux/lockdep.h
+++ b/include/linux/lockdep.h
@@ -339,10 +339,21 @@ static inline int lockdep_match_key(struct lockdep_map *lock,
*
* Values for "read":
*
* 0: exclusive (write) acquire
* 1: read-acquire (no recursion allowed)
* 2: read-acquire with same-instance recursion allowed
*
+ * Note that we have an assumption that a lock class cannot ever be both
+ * read and recursive-read.
+ */
+enum lock_type {
+ LOCK_TYPE_EXCLUSIVE = 0,
+ LOCK_TYPE_WRITE = 0,
+ LOCK_TYPE_READ,
+ LOCK_TYPE_RECURSIVE,
+};
+
+/*
* Values for check:
*
* 0: simple checks (freeing, held-at-exit-time, etc.)
@@ -588,9 +599,14 @@ static inline void print_irqtrace_events(struct task_struct *curr)
* on the per lock-class debug mode:
*/

-#define lock_acquire_exclusive(l, s, t, n, i) lock_acquire(l, s, t, 0, 1, n, i)
-#define lock_acquire_shared(l, s, t, n, i) lock_acquire(l, s, t, 1, 1, n, i)
-#define lock_acquire_shared_recursive(l, s, t, n, i) lock_acquire(l, s, t, 2, 1, n, i)
+#define lock_acquire_exclusive(l, s, t, n, i) \
+ lock_acquire(l, s, t, LOCK_TYPE_EXCLUSIVE, 1, n, i)
+
+#define lock_acquire_shared(l, s, t, n, i) \
+ lock_acquire(l, s, t, LOCK_TYPE_READ, 1, n, i)
+
+#define lock_acquire_shared_recursive(l, s, t, n, i) \
+ lock_acquire(l, s, t, LOCK_TYPE_RECURSIVE, 1, n, i)

#define spin_acquire(l, s, t, i) lock_acquire_exclusive(l, s, t, NULL, i)
#define spin_acquire_nest(l, s, t, n, i) lock_acquire_exclusive(l, s, t, n, i)
@@ -339,10 +339,21 @@ static inline int lockdep_match_key(struct lockdep_map *lock,
* Values for "read":
*
- * 0: exclusive (write) acquire
- * 1: read-acquire (no recursion allowed)
- * 2: read-acquire with same-instance recursion allowed
+ * LOCK_TYPE_EXCLUSIVE (LOCK_TYPE_WRITE): exclusive (write) acquire
+ * LOCK_TYPE_READ: read-acquire (no recursion allowed)
+ * LOCK_TYPE_RECURSIVE: read-acquire with same-instance recursion allowed
*
--
1.8.3.1


2019-05-13 13:34:07

by Peter Zijlstra

[permalink] [raw]
Subject: Re: [PATCH 01/17] locking/lockdep: Add lock type enum to explicitly specify read or write locks

On Mon, May 13, 2019 at 05:11:47PM +0800, Yuyang Du wrote:
> + * Note that we have an assumption that a lock class cannot ever be both
> + * read and recursive-read.

We have such locks in the kernel... see:

kernel/qrwlock.c:queued_read_lock_slowpath()

And yes, that is somewhat unfortunate, but hard to get rid of due to
hysterical raisins.

2019-05-14 01:33:16

by Yuyang Du

[permalink] [raw]
Subject: Re: [PATCH 01/17] locking/lockdep: Add lock type enum to explicitly specify read or write locks

Thanks for review.

On Mon, 13 May 2019 at 19:45, Peter Zijlstra <[email protected]> wrote:
>
> On Mon, May 13, 2019 at 05:11:47PM +0800, Yuyang Du wrote:
> > + * Note that we have an assumption that a lock class cannot ever be both
> > + * read and recursive-read.
>
> We have such locks in the kernel... see:
>
> kernel/qrwlock.c:queued_read_lock_slowpath()
>
> And yes, that is somewhat unfortunate, but hard to get rid of due to
> hysterical raisins.

That is ok, then LOCK_TYPE_RECURSIVE has to be 3 such that
LOCK_TYPE_RECURSIVE & LOCK_TYPE_READ != 0. I thought to do this in the
first place without assuming. Anyway, it is better to know.

And I guess in a task:

(1) read(X);
recursive_read(x); /* this is ok ? */

(2) recursive_read(x);
read(x) /* not ok ? */

Either way, very small change may need to be made.

2019-05-14 12:07:35

by Peter Zijlstra

[permalink] [raw]
Subject: Re: [PATCH 01/17] locking/lockdep: Add lock type enum to explicitly specify read or write locks

On Tue, May 14, 2019 at 09:31:48AM +0800, Yuyang Du wrote:
> Thanks for review.
>
> On Mon, 13 May 2019 at 19:45, Peter Zijlstra <[email protected]> wrote:
> >
> > On Mon, May 13, 2019 at 05:11:47PM +0800, Yuyang Du wrote:
> > > + * Note that we have an assumption that a lock class cannot ever be both
> > > + * read and recursive-read.
> >
> > We have such locks in the kernel... see:
> >
> > kernel/qrwlock.c:queued_read_lock_slowpath()
> >
> > And yes, that is somewhat unfortunate, but hard to get rid of due to
> > hysterical raisins.
>
> That is ok, then LOCK_TYPE_RECURSIVE has to be 3 such that
> LOCK_TYPE_RECURSIVE & LOCK_TYPE_READ != 0. I thought to do this in the
> first place without assuming. Anyway, it is better to know.
>
> And I guess in a task:
>
> (1) read(X);
> recursive_read(x); /* this is ok ? */

Correct, that is the use-case for that 'funny' construct.

> (2) recursive_read(x);
> read(x) /* not ok ? */

Indeed, read can block due to a pending writer, while recurise_read will
not suffer like that.

> Either way, very small change may need to be made.

OK, excellent.